from django.contrib.auth.decorators import login_required
from django.contrib.auth.models import User
from django.db.models import Count, Q, Case, When, IntegerField, Value
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse

from core.custom_annotations import group_required, in_groups
from core.models import Feedback, Submission, SubmissionType


@login_required(login_url='/')
def user_home(request):
    if in_groups(request.user, ('Students', )):
        return student_view(request)
    elif in_groups(request.user, ('Tutors', )):
        return tutor_view(request)
    elif in_groups(request.user, ('Reviewers', )):
        return reviewer_view(request)
    else:
        return HttpResponseRedirect(reverse('index'))

def get_annotated_feedback_count():
    """ Annotates submission lists with counts

    count both
        * number of submission per submission type
        * count of received feedback per submission type
        *
    Alternative with case
        Count(Case(
            When(submissions__feedback_list__origin=Feedback.MANUAL,
                then=Value(1)), output_field=IntegerField())
        )

    Returns:
        annotated queryset
    """
    return SubmissionType.objects\
        .annotate( # to display only manual
            feedback_count=Count(
                Case(
                    When(Q(submissions__feedback__isnull=False) & Q(submissions__feedback__status=Feedback.ACCEPTED),
                        then=Value(1)), output_field=IntegerField(),
                )
            )
        ).annotate(
            submission_count=Count('submissions')
        ).all().order_by('name')

@group_required('Tutors')
def tutor_view(request):
    context = {
        'submission_type_list': get_annotated_feedback_count(),
        'feedback_list': Feedback.objects.filter(of_tutor=request.user),
        'feedback_open_list': Feedback.objects.filter(Q(status=Feedback.OPEN) & ~Q(of_tutor=request.user)),
    }
    return render(request, 'core/tutor_startpage.html', context)


@group_required('Students')
def student_view(request):
    context = {
        'student': request.user,
        'submission_list': Submission.objects.filter(student__user=request.user)
    }
    return render(request, 'core/student_startpage.html', context)


@group_required('Reviewers')
def reviewer_view(request):
    context = {
        'submission_type_list': get_annotated_feedback_count(),
        'tutor_list': User.objects.annotate(Count('corrected_submissions')).filter(groups__name='Tutors')
                                                                           .order_by('-corrected_submissions__count'),
        'feedback_list_manual': Feedback.objects.filter(origin=Feedback.MANUAL),
        'feedback_list_empty': Feedback.objects.filter(origin=Feedback.WAS_EMPTY),
        'feedback_list_did_not_compile': Feedback.objects.filter(origin=Feedback.DID_NOT_COMPILE),
        'feedback_list_could_not_link': Feedback.objects.filter(origin=Feedback.COULD_NOT_LINK),
    }
    return render(request, 'core/reviewer_startpage.html', context)
